"""
Optimizers using other packages. Currently we support optimizers in scipy.
"""

import numpy as np

from tqdm.auto import tqdm
from scipy import optimize

from .. import models

__all__ = ["scipy_optimzer", "scipy_L_BFGS_B"]


class scipy_optimizer(object):
    def __init__(self) -> None:
        pass

    def run(self, model, method=None, diff_mode="manual", tol=None, kwargs_minimize={}):
        diff_mode = diff_mode.lower()
        assert diff_mode in ["manual", "num"]

        if type(model) is models.GWR_Ridge:
            x0 = np.concatenate(
                [np.array(model.kernel.params), np.array([model.penalty])]
            )
        elif type(model) is models.GWR or type(model) is models.ScaGWR:
            x0 = np.array(model.kernel.params)
        else:
            raise ValueError("Unknown model class")

        x0 = model._to_unconstrained(x0)

        def f(x):
            return model.unconstrained_loss(x)

        g = None
        if diff_mode == "manual":

            def g(x):
                return model.unconstrained_grad(x)

        res = optimize.minimize(
            fun=f, x0=x0, method=method, jac=g, tol=tol, **kwargs_minimize
        )

        model.set_params(res.x)

        return res

    def run_scalar(
        self,
        model,
        method="brent",
        bracket=None,
        bounds=None,
        tol=None,
        options=None,
        aicc=False,
        kwargs_minimize=dict(),
    ):
        if aicc:

            def f(x):
                z = model._to_constrained(np.array([x]))
                return model.AICc([z])

        else:

            def f(x):
                return model.unconstrained_loss(np.array([x]))

        res = optimize.minimize_scalar(
            fun=f,
            bracket=bracket,
            bounds=bounds,
            method=method,
            tol=tol,
            options=options,
            **kwargs_minimize,
        )

        model.set_params(np.array([res.x]))

        return res


class scipy_L_BFGS_B(object):
    """same setting to the 'scgwr' package in R
    see:  https://github.com/cran/scgwr/blob/master/R/scgwr.R
    """

    def __init__(self) -> None:
        self.xla_jit = True

    def run(self, model, diff_mode="manual", tol=None, kwargs_minimize=dict()):

        diff_mode = diff_mode.lower()
        assert diff_mode in ["manual", "num"]

        if type(model) is models.GWR_Ridge:
            x0 = np.concatenate(
                [np.array(model.kernel.params), np.array([model.penalty])]
            )

            def f(x):
                return model.loocv_loss(x[:-1], x[-1])

        elif type(model) is models.GWR or type(model) is models.ScaGWR:
            x0 = np.array(model.kernel.params)

            def f(x):
                return model.loocv_loss(x)

        else:
            raise ValueError("Unknown model class")

        def f64(x):
            return np.array(f(x), dtype=np.float64)

        g = None
        if diff_mode == "manual":
            if type(model) is models.GWR_Ridge:

                def g(x):
                    g1 = model.grad_params_loocv(x[:-1], x[-1])
                    g2 = model.grad_penalty_loocv(x[:-1], x[-1])

                    return np.concatnate([g1, np.array([g2])])

            elif type(model) is models.GWR or type(model) is models.ScaGWR:

                def g(x):
                    return model.grad_params_loocv(x)

        if g is None:
            g64 = None
        else:

            def g64(x):
                return np.array(g(x)).astype(np.float64)

        res = optimize.minimize(
            fun=f64,
            x0=np.array(x0, dtype=np.float64),
            method="L-BFGS-B",
            jac=g64,
            bounds=optimize.Bounds(lb=np.zeros(len(x0)), ub=np.inf * np.ones(len(x0))),
            tol=tol,
            **kwargs_minimize,
        )

        model.set_params(res.x, transform=False)

        return res




"""
Optimizers using stochastic gradients.
"""
from jax import numpy as jnp
from jax import grad, value_and_grad, random
from jax import jit, lax

from tqdm.auto import tqdm

from .. import models

__all__ = ["SGD", "ASGD", "SGDarmijo", "Adam", "Yogi"]


class SGD(object):
    """
    Stochastic Gradient Descent Algorithm

    reference:
    Bottou, L. (2010). Large-scale machine learning with stochastic gradient descent.
    Proceedings of COMPSTAT 2010 - 19th International Conference on Computational Statistics,
    Keynote, Invited and Contributed Papers, 177–186. https://doi.org/10.1007/978-3-7908-2604-3_16
    """

    def __init__(self, learning_rate0=1.0, lam=1e-4):
        self.learning_rate0 = float(learning_rate0)
        self.lam = float(lam)

        self.xla_jit = True

    def lr_schedule(self, t):
        return self.learning_rate0 / (1 + self.lam * self.learning_rate0 * t)

    def run(
        self,
        model,
        maxiter=1000,
        batchsize=100,
        PRNGkey=random.PRNGKey(seed=123),
        diff_mode="manual",
        tol=0.001,
        n_iter_no_change=100,
        verbose=True,
    ):
        maxiter = int(maxiter)
        batchsize = int(batchsize)
        assert batchsize <= model.N

        self.batchsize = batchsize
        self.N = model.N

        x0, [f, g, f_and_g] = self._init_optimizer(model, diff_mode)

        key, PRNGkey = random.split(PRNGkey)
        idx = random.choice(key, model.N, shape=(batchsize,), replace=True)
        loss = [float(f(x0, idx=idx))]
        x = x0
        best_loss = float(loss[-1])
        count = 0
        self.lr_log = []
        with tqdm(total=maxiter, disable=not verbose) as pbar:
            for t in range(maxiter):
                pbar.update(1)
                pbar.set_description("loss={}".format(loss[-1]))

                self.lr = self.lr_schedule(t + 1)
                key, PRNGkey = random.split(PRNGkey)
                idx = random.choice(key, model.N, shape=(batchsize,), replace=True)
                x, l = self.step(t + 1, x, f, g, f_and_g, idx)
                self.lr_log.append(float(self.lr))

                loss.append(l)

                # convergence check
                if l - best_loss < tol:
                    best_loss = float(min(loss))
                    count = 0
                else:
                    count += 1
                    if count >= n_iter_no_change:
                        self.converged = True
                        break
            else:
                self.converged = False

        model.set_params(x)

        return loss

    def _init_optimizer(self, model, diff_mode):
        diff_mode = diff_mode.lower()
        assert diff_mode in ["manual", "auto"]

        if type(model) is models.GWR_Ridge:
            x0 = jnp.concatenate(
                [jnp.array(model.kernel.params), jnp.array([model.penalty])]
            )

        elif type(model) is models.GWR or type(model) is models.ScaGWR:
            x0 = jnp.array(model.kernel.params)
        else:
            raise ValueError("Unknown model class")

        x0 = model._to_unconstrained(x0)

        def f(x, idx):
            return model.unconstrained_loss(x, idx)

        if diff_mode == "manual":

            def g(x, idx):
                return model.unconstrained_grad(x, idx)

            def f_and_g(x, idx):
                return (
                    model.unconstrained_loss(x, idx),
                    model.unconstrained_grad(x, idx),
                )

        elif diff_mode == "auto":

            def g(x, idx):
                return grad(model.unconstrained_loss, argnums=0)(x, idx)

            def f_and_g(x, idx):
                return value_and_grad(model.unconstrained_loss, argnums=0)(x, idx)

        if self.xla_jit:
            f, g, f_and_g = jit(f), jit(g), jit(f_and_g)

        return x0, [f, g, f_and_g]

    def step(self, t, x, f, g, f_and_g, idx):
        grads = g(x, idx)
        x_new = x - self.lr * grads
        loss = f(x, idx)
        return x_new, float(loss)


class ASGD(SGD):
    """
    Avereaged Stochastic Gradient Descent Algorithm

    reference:
    Bottou, L. (2010). Large-scale machine learning with stochastic gradient descent.
    Proceedings of COMPSTAT 2010 - 19th International Conference on Computational Statistics,
    Keynote, Invited and Contributed Papers, 177–186. https://doi.org/10.1007/978-3-7908-2604-3_16
    """

    def __init__(self, learning_rate0=1.0, lam=1e-4):
        super().__init__(learning_rate0=learning_rate0, lam=lam)

    def lr_schedule(self, t):
        return self.learning_rate0 * (1 + self.lam * self.learning_rate0 * t) ** (-0.75)

    def step(self, t, x, f, g, f_and_g, idx):
        grads = g(self._x_sgd, idx)
        self._x_sgd = self._x_sgd - self.lr * grads
        x_new = (t * x + self._x_sgd) / (t + 1)
        loss = f(x_new, idx)
        return x_new, float(loss)

    def _init_optimizer(self, model, diff_mode):
        x0, [f, g, f_and_g] = super()._init_optimizer(model, diff_mode)
        self._x_sgd = jnp.array(x0)
        return x0, [f, g, f_and_g]


class SGDarmijo(SGD):
    """
    Stochastic Gradient Descent Algorithm with Armijo Line-search

    reference:
    Vaswani, S., Mishkin, A., Laradji, I., Schmidt, M., Gidel, G., & Lacoste-Julien, S. (2019).
    Painless stochastic gradient: Interpolation, line-search, and convergence rates.
    Advances in neural information processing systems, 32.
    """

    def __init__(
        self,
        learning_rate0=1.0,
        c=0.5,
        ls_decay=0.5,
        reset_decay=2.0,
        search_from_lr0=False,
    ):
        self.learning_rate0 = float(learning_rate0)

        assert c > 0.0
        self.c = float(c)

        assert 0.0 < ls_decay and ls_decay < 1.0
        self.ls_decay = float(ls_decay)

        assert reset_decay >= 1.0
        self.reset_decay = float(reset_decay)

        self.search_from_lr0 = search_from_lr0

        self.lr = float(learning_rate0)

        self._ls_max = 100

        self.xla_jit = True

    def lr_schedule(self, t):
        # set initial lr on line search
        if self.search_from_lr0:
            lr = self.learning_rate0
        else:
            lr = self.lr * self.reset_decay ** (self.batchsize / self.N)
            # print(f"lr0 = {lr}")
        return lr  # / self.ls_decay

    def step(self, t, x, f, g, f_and_g, idx):
        grads = g(x, idx)
        loss = f(x, idx)
        # line search
        self.lr = self.armijo_search(x, grads, self.lr, f=lambda x: f(x, idx))
        x_new = x - self.lr * grads
        return x_new, float(loss)

    def armijo_search(self, x, grads, lr, f):
        g2 = jnp.sum(jnp.square(grads))
        f0 = f(x)
        cond_fun = lambda lr: not self.armijo_cond(f, f0, x, grads, g2, lr)
        body_fun = lambda lr: lr * self.ls_decay
        # lr_searched = lax.while_loop(cond_fun=cond_fun, body_fun=body_fun, init_val=lr)

        # print(f"start line search (lr={lr})")
        i = 0
        while cond_fun(lr) or i > self._ls_max:
            # print(f"lr={lr}")
            lr = body_fun(lr)
            i += 1

        lr_searched = lr

        return lr_searched

    def armijo_cond(self, f, f0, x, grads, g2, lr):
        x_new = x - lr * grads
        f_new = f(x_new)
        return f_new <= f0 - self.c * lr * g2


class Adam(SGD):
    def __init__(
        self, learning_rate0=1.0, b1=0.9, b2=0.999, eps=1e-8, correct_bias=True
    ):
        assert 0 <= b1 < 1
        self.b1 = float(b1)
        assert 0 <= b2 < 1
        self.b2 = float(b2)
        self.eps = float(eps)
        self.correct_bias = correct_bias
        super().__init__(learning_rate0, lam=0)

    def _init_optimizer(self, model, diff_mode):
        x0, [f, g, f_and_g] = super()._init_optimizer(model, diff_mode)
        self.m = jnp.zeros_like(x0)  # 1st moment
        self.v = jnp.zeros_like(x0)  # 2nd moment
        self.lr = float(self.learning_rate0)
        return x0, [f, g, f_and_g]

    def step(self, t, x, f, g, f_and_g, idx):
        grads = g(x, idx)
        g2 = jnp.square(grads)

        # update moments
        self.m = self.b1 * self.m + (1 - self.b1) * grads
        self.v = self.b2 * self.v + (1 - self.b2) * g2

        # moments' bias-correction
        if self.correct_bias:
            m = self.m / (1 - self.b1 ** (t + 1))
            v = self.v / (1 - self.b2 ** (t + 1))
        else:
            m = self.m
            v = self.v

        x_new = x - self.lr * m / (jnp.sqrt(v) + self.eps)
        loss = f(x, idx)
        return x_new, float(loss)


class Yogi(Adam):
    def step(self, t, x, f, g, f_and_g, idx):
        grads = g(x, idx)
        g2 = jnp.square(grads)

        # update moments
        self.m = self.b1 * self.m + (1 - self.b1) * grads
        self.v = self.v - (1 - self.b2) * jnp.sign(self.v - g2) * g2

        x_new = x - self.lr * self.m / (jnp.sqrt(self.v) + self.eps)
        loss = f(x, idx)
        return x_new, float(loss)




try:
    from jax import numpy as jnp

    def _mask_loocv(w, i):
        if i is None:
            return w
        else:
            return w.at[i].set(0.0)

except:
    import numpy as jnp

    def _mask_loocv(w, i):
        if i is None:
            return w
        else:
            w[i] = 0.0
            return w

    def vmap(f):
        def vf(idx):
            return jnp.stack(list(map(f, idx)), axis=0)

        return vf


from scipy.spatial import KDTree

__all__ = [
    "Gaussian",
    "Exponential",
    "Epanechnikov",
    "Triangular",
    "Biweight",
    "LinearMultiscale",
    "stGaussian",
    "stExponential",
    "stEpanechnikov",
    "stTriangular",
    "stBiweight",
]


class _baseKernel(object):
    def __init__(self, params):
        self.params = params

    def k(self, x1, x2, params):
        raise NotImplementedError()

    def dk(self, x1, x2, params):
        raise NotImplementedError()

    def __call__(self, x1, x2=None, params=None, loocv=None):
        return self.forward(x1, x2, params, loocv)

    def forward(self, x1, x2=None, params=None, loocv=None):
        if x2 is None:
            x2 = x1.reshape(1, -1)
        if params is None:
            params = self.params

        return _mask_loocv(self.k(x1, x2, params), loocv)
        # if loocv:
        #     return jnp.where(jnp.all(x1 == x2, axis=1), 0, self.k(x1, x2, params))
        # else:
        #     return self.k(x1, x2, params)

    def grad(self, x1, x2=None, params=None, loocv=None):
        if x2 is None:
            x2 = x1.reshape(1, -1)
        if params is None:
            params = self.params

        return _mask_loocv(self.dk(x1, x2, params), loocv)
        # if loocv:
        #     return jnp.where(
        #         jnp.all(x1 == x2, axis=1)[:, None], 0, self.dk(x1, x2, params)
        #     )
        # else:
        #     return self.dk(x1, x2, params)


class _scaledKernel(_baseKernel):
    def _dist(self, x1, x2, params):
        d = jnp.linalg.norm(x1 - x2, axis=-1)
        return params[0] * d, d

    def _scaledk(self, d):
        raise NotImplementedError()

    def _scaleddk(self, d):
        raise NotImplementedError()

    def k(self, x1, x2, params):
        scaled, _ = self._dist(x1, x2, params)
        return self._scaledk(scaled)

    def dk(self, x1, x2, params):
        scaled, d = self._dist(x1, x2, params)
        return (self._scaleddk(scaled) * d).reshape(-1, 1)

    def _scaled_k_inv(self, k):
        raise NotImplementedError()

    def init_param(self, sites, idx=None):
        if idx is None:
            idx = jnp.arange(len(sites), dtype=int)

        # apply map to avoid making N*N matrix
        med_dists = list(
            map(lambda i: jnp.median(self._dist(sites[i], sites, [1.0])[1]), idx)
        )
        d_median = jnp.median(jnp.array(med_dists))

        # search param that satisfy 0.5 * k(param0 * d_median) = k(0)
        # -> param0 = k^{-1}(0.5 * k(0)) / d_median
        k0 = self._scaledk(0.0)
        self.params = [self._scaled_k_inv(0.5 * k0) / d_median]


class Gaussian(_scaledKernel):
    def _scaledk(self, d):
        return jnp.exp(-jnp.square(d))

    def _scaleddk(self, d):
        return -2 * d * self._scaledk(d)

    def _scaled_k_inv(self, k):
        return jnp.sqrt(-jnp.log(k))


class Exponential(_scaledKernel):
    def _scaledk(self, d):
        return jnp.exp(-d)

    def _scaleddk(self, d):
        return -self._scaledk(d)

    def _scaled_k_inv(self, k):
        return -jnp.log(k)


class Epanechnikov(_scaledKernel):
    def _scaledk(self, d):
        return jnp.where(d <= 1, 1 - jnp.square(d), 0)

    def _scaleddk(self, d):
        return jnp.where(d <= 1, -2 * d, 0)

    def _scaled_k_inv(self, k):
        return jnp.sqrt(1 - k)


class Triangular(_scaledKernel):
    def _scaledk(self, d):
        return jnp.where(d <= 1, 1 - d, 0)

    def _scaleddk(self, d):
        return jnp.where(d <= 1, -1, 0)

    def _scaled_k_inv(self, k):
        return 1 - k


class Biweight(_scaledKernel):
    def _scaledk(self, d):
        return jnp.where(d <= 1, jnp.square(1 - jnp.square(d)), 0)

    def _scaleddk(self, d):
        return jnp.where(d <= 1, -4 * d * (1 - jnp.square(d)), 0)

    def _scaled_k_inv(self, k):
        return jnp.sqrt(1 - jnp.sqrt(k))


class _KDTreeKernel(_baseKernel):
    def __init__(self, sites, params):
        super().__init__(params)
        self._kdtree = KDTree(sites)

    def _knn(self, sites, k=1):
        dist, idx = self._kdtree.query(sites, k=k)
        return dist, idx


class AdaptiveKernel(_baseKernel):
    """Adaptive Kernel"""

    def __init__(self, params, k_dist=Biweight([1.0])):
        super().__init__(params)

        if isinstance(k_dist, _scaledKernel):
            self.k_dist = k_dist._scaledk
        elif callable(k_dist):
            self.k_dist = k_dist
        else:
            raise ValueError("Unknown k_dist")

    def k(self, x1, x2, params):
        k = int(params[0])
        d = jnp.linalg.norm(x1 - x2, axis=-1)
        bw = jnp.partition(d, kth=k - 1)[k - 1] * (1 + 1e-7)

        return self.k_dist(d / bw)


class LinearMultiscale(_KDTreeKernel):
    def __init__(
        self,
        sites,
        params=jnp.array([0.01, 1]) ** 2,
        base_kernel=None,
        n_poly=4,
        n_neighbour=100,
    ):
        # default settings are the same to that of 'scgwr' package in R
        # see:  https://github.com/cran/scgwr/blob/master/R/scgwr.R
        super().__init__(sites, params)
        assert n_poly >= 1
        self.n_poly = int(n_poly)
        assert n_neighbour >= 1
        self.n_neighbour = int(n_neighbour)

        self.sites = sites

        if base_kernel is None:

            knn_dist, _ = self._knn(sites, k=self.n_neighbour)
            ban0 = jnp.median(
                knn_dist[:, min(50 + 1, self.n_neighbour + 1)]
            ) / jnp.sqrt(3)
            self.base_kernel = Gaussian([1 / ban0])
        else:
            self.base_kernel = base_kernel
        self.N = len(self.sites)

    def k(self, x1, x2, params):
        a = params[0]
        b = params[1:]
        w0 = self.base_kernel(x1, x2, loocv=None)
        # assume constantly decreasing kernel
        # thres = jnp.partition(w0, self.n_neighbour)[self.n_neighbour - 1] # jnp.partition is not implemented yet
        thres = jnp.sort(w0)[::-1][self.n_neighbour - 1]
        w0 = jnp.where(w0 <= thres, 0, w0)

        p = jnp.arange(1, self.n_poly + 1, dtype=int)
        if len(jnp.array(b)) == 1:
            b = b**p
        Wp = w0[None] ** (4 / jnp.power(2, p))[:, None]
        return a + jnp.sum(b[:, None] * Wp, axis=0)

    def dk(self, x1, x2, params):
        raise NotImplementedError()


class _scaledSTKernel(_scaledKernel):
    """STkernel (space-time kernel) assumes three dimentional coordinates.
    First two coordinates are location (x,y) coordinates, whearas the last third coordinate is the temporal time-stamp t
    Space-time distance is determined as follows:
        [Space-time distance] = params[0] * [spatial distance] + params[1] * [temporal distance]
    """

    def _dist(self, x1, x2, params):
        vec = x1 - x2

        d = jnp.linalg.norm(vec[:, :2], axis=-1)
        t = jnp.abs(vec[:, 2])
        return params[0] * d + params[1] * t, d, t

    def k(self, x1, x2, params):
        scaled, _, _ = self._dist(x1, x2, params)
        return self._scaledk(scaled)

    def dk(self, x1, x2, params):
        scaled, d, t = self._dist(x1, x2, params)
        dkdx = self._scaleddk(scaled)
        return jnp.stack([dkdx * d, dkdx * t], axis=1)

    def _scaled_k_inv(self, k):
        raise NotImplementedError()

    def init_param(self, sites, idx=None):
        if idx is None:
            idx = jnp.arange(len(sites), dtype=int)

        # apply map to avoid making N*N matrix
        med_dists = []
        med_ts = []
        for i in idx:
            tmp = self._dist(sites[i], sites, [1.0, 1.0])
            med_dists.append(jnp.median(tmp[1]))
            med_ts.append(jnp.median(tmp[2]))

        d_median = jnp.median(jnp.array(med_dists))
        t_median = jnp.median(jnp.array(med_ts))

        dt_scale = d_median / t_median
        # search param that satisfy 0.5 * k(param0 * 2 * d_median) = k(0)
        # -> param0 = k^{-1}(0.5 * k(0)) / d_median / 2
        k0 = self._scaledk(0.0)
        param0 = self._scaled_k_inv(0.5 * k0) / d_median / 2

        # param0 * 2 * d_median = param0 * (d_median + dt_scale * t_median)
        # -> param1 = param0 * dt_scale
        param1 = param0 * dt_scale

        self.params = [param0, param1]


class stGaussian(_scaledSTKernel):
    def _scaledk(self, d):
        return jnp.exp(-jnp.square(d))

    def _scaleddk(self, d):
        return -2 * d * self._scaledk(d)

    def _scaled_k_inv(self, k):
        return jnp.sqrt(-jnp.log(k))


class stExponential(_scaledSTKernel):
    def _scaledk(self, d):
        return jnp.exp(-d)

    def _scaleddk(self, d):
        return -self._scaledk(d)

    def _scaled_k_inv(self, k):
        return -jnp.log(k)


class stEpanechnikov(_scaledSTKernel):
    def _scaledk(self, d):
        return jnp.where(d <= 1, 1 - jnp.square(d), 0)

    def _scaleddk(self, d):
        return jnp.where(d <= 1, -2 * d, 0)

    def _scaled_k_inv(self, k):
        return jnp.sqrt(1 - k)


class stTriangular(_scaledSTKernel):
    def _scaledk(self, d):
        return jnp.where(d <= 1, 1 - d, 0)

    def _scaleddk(self, d):
        return jnp.where(d <= 1, -1, 0)

    def _scaled_k_inv(self, k):
        return 1 - k


class stBiweight(_scaledSTKernel):
    def _scaledk(self, d):
        return jnp.where(d <= 1, jnp.square(1 - jnp.square(d)), 0)

    def _scaleddk(self, d):
        return jnp.where(d <= 1, -4 * d * (1 - jnp.square(d)), 0)

    def _scaled_k_inv(self, k):
        return jnp.sqrt(1 - jnp.sqrt(k))


class SpectralMixture(_baseKernel):
    """Spectral Mixture (SM) kernel for Gaussian Process

    inspired by

    Andrew Wilson, Ryan Adams
    Gaussian Process Kernels for Pattern Discovery and Extrapolation
    Proceedings of the 30th International Conference on Machine Learning, PMLR
    28(3):1067-1075, 2013.
    http://proceedings.mlr.press/v28/wilson13.html
    """

    def __init__(self, params):
        super().__init__(params)

        if len(self.params) % 3 != 0:
            raise ValueError("number of parameters must be (3 * number of mixture)")

        self.n_mixture = len(self.params) / 3

    def _dist(self, x1, x2):
        d = jnp.linalg.norm(x1 - x2, axis=-1)
        return d

    def _parse_params(self, params):
        mat = jnp.array(params).reshape((3, self.n_mixture))
        w, v, mu = mat[0], mat[1], mat[2]
        return w, v, mu

    def k(self, x1, x2, params):
        w, v, mu = self._parse_params(params)
        d = self._dist(x1, x2)

        e, c = self._get_ec(d, v, mu)
        return jnp.sum(w * e * c)

    def dk(self, x1, x2, params):
        w, v, mu = self._parse_params(params)
        d = self._dist(x1, x2)

        e, c = self._get_ec(d, v, mu)

        dw = e * c
        dv = w * c * e * (-2 * jnp.square(jnp.pi * d))
        dmu = w * e * (-2 * jnp.pi * d) * jnp.sin(2 * jnp.pi * d * mu)

        return jnp.concatenate([dw, dv, dw]).reshape(-1, 1)

    def _get_ec(self, d, v, mu):
        e = jnp.exp(-2 * jnp.square(jnp.pi * d) * v)
        c = jnp.cos(2 * jnp.pi * d * mu)

        return e, c



from .. import models, kernels

try:
    from .existings import optax_optimizer, scipy_optimizer, scipy_L_BFGS_B
    from .sg import SGD, ASGD, SGDarmijo, Adam, Yogi
    from .vr import SVRG, KatyushaXs, KatyushaXw
    from .second import SGN, SGN_BFGS, SGN_LM
    from .golden import golden_section

    __all__ = [
        "optax_optimizer",
        "scipy_optimzer",
        "scipy_L_BFGS_B",
        "SGD",
        "ASGD",
        "SGDarmijo",
        "Adam",
        "Yogi",
        "SVRG",
        "KatyushaXs",
        "KatyushaXw",
        "SGN",
        "SGN_BFGS",
        "SGN_LM",
        "golden_section",
    ]

except:
    from .existings_numpy import scipy_optimizer, scipy_L_BFGS_B
    from .sg_numpy import SGD, ASGD, SGDarmijo
    from .golden import golden_section

    print(
        "JAX is not available. Installing JAX is strongly recommended for performance."
    )

    __all__ = [
        "SGD",
        "ASGD",
        "SGDarmijo",
        "scipy_optimzer",
        "scipy_L_BFGS_B",
        "golden_section",
    ]



"""
Golden Section Search for adaptive bandwidth 
"""

import numpy as np

from tqdm.auto import tqdm
from .. import models

__all__ = ["golden_section"]


class golden_section(object):
    def __init__(self) -> None:
        super().__init__()

    def run(
        self, model, maxiter=1000, bracket=None, tol=1e-5, aicc=False, verbose=True
    ):
        assert isinstance(model, models.GWR)

        if bracket is None:

            a, c = model.D + 1, model.N
        else:
            a, c = bracket
            assert (type(a) is int) and (type(c) is int)
            if a > c:
                a, c = c, a

        if aicc:

            def f(x):
                return model.AICc(np.array([x]))

        else:

            def f(x):
                return model.loocv_loss(np.array([x]))
                # return self._loocv(x, model)

        fa, fc = f(a), f(c)

        # check interval
        if a == c:
            opt = a
            self.loss = [fa]
        elif a + 1 == c:
            if fa < fc:
                self.loss = [fa]
                opt = a
            else:
                self.loss = [fc]
                opt = c

        delta = (1 + np.sqrt(5)) / 2 - 1

        # make triplet [a, b, c]
        b = self._propose_new(a, c, delta)
        fb = f(b)

        self.loss = [float(fb)]
        with tqdm(total=maxiter, disable=not verbose) as pbar:
            for i in range(maxiter):
                pbar.update(1)
                pbar.set_description(
                    "loss = {:.2f}, trpilet={}".format(self.loss[-1], (a, b, c))
                )

                # fmt: off
                # do not format code between fmt: off and fmt: on
                if b - a >= c - b:
                    # propose new point in [a, b]
                    d = self._propose_new(a, b, delta)
                    fd = f(d)

                    # compare f(d) with center value f(b) and get new triplet
                    if fd < fb:
                        # next triplet is [a, d, b]
                        a, b, c = a, d, b
                        fa, fb, fc = fa, fd, fb
                    else:
                        # next triplet is [d, b, c]
                        a, b, c, = d, b, c
                        fa, fb, fc, = fd, fb, fc
                else:
                    # propose new point in [b, c]
                    d = self._propose_new(b, c, delta)
                    fd = f(d)

                    # compare f(d) with center value f(b) and get new triplet
                    if fd < fb:
                        # next triplet is [b, d, c]
                        a, b, c, =  b, d, c
                        fa, fb, fc, = fb, fd, fc
                    else:
                        # next triplet is [a, b, d]
                        a, b, c = a, b, d
                        fa, fb, fc = fa, fb, fd
                # fmt: on

                if not (a < b and b < c):
                    raise ValueError(f"the triplet {[a, b ,c]} is not in order.")

                # check convergence
                idx = np.argmin([fa, fb, fc])
                opt = [a, b, c][idx]
                self.loss.append(float([fa, fb, fc][idx]))
                if max(fa, fb, fc) - min(fa, fb, fc) < tol or a + 2 >= c:
                    pbar.set_description(
                        "loss = {:.2f}, trpilet={}".format(self.loss[-1], (a, b, c))
                    )
                    break
            else:
                print(
                    "Caution: golden section search reached maximum number of iteration. maxiter may not enough large."
                )

        model.kernel.params = [opt]
        return self.loss

    def _propose_new(self, a, c, delta):
        b = delta * a + (1 - delta) * c
        b_floor, b_ceil = np.floor(b), np.ceil(b)

        if b_floor != a and b_ceil != c:
            b = np.round(b)
        elif b_floor == a:
            b = b_ceil
        elif b_ceil == c:
            b = b_floor

        return int(b)



"""
Optimizers using other packages. Currently we support optimizers in scipy and optax.
"""

import numpy as np
from jax import numpy as jnp
from jax import grad, value_and_grad, random, jit
import optax

from tqdm.auto import tqdm
from scipy import optimize

from .. import models

__all__ = ["optax_optimizer", "scipy_optimzer", "scipy_L_BFGS_B"]


class optax_optimizer(object):
    def __init__(
        self, optax_optim=optax.sgd(lambda count: max(1e-5, 1 / (1 + count)))
    ) -> None:
        super().__init__()
        self.optax_optim = optax_optim

    def run(
        self,
        model,
        maxiter=1000,
        batchsize=100,
        PRNGkey=None,
        diff_mode="manual",
        tol=0.001,
        n_iter_no_change=100,
        verbose=True,
    ):
        diff_mode = diff_mode.lower()
        assert diff_mode in ["manual", "auto"]

        if batchsize is None:
            batchsize = model.N
        else:
            if PRNGkey is None:
                if batchsize != model.N:
                    raise ValueError("jax random.PRNGkey should be specified")
                else:
                    print("Batch learning mode")
            assert batchsize <= model.N

        if type(model) is models.GWR_Ridge:
            x0 = jnp.concatenate(
                [jnp.array(model.kernel.params), jnp.array([model.penalty])]
            )
        elif type(model) is models.GWR or type(model) is models.ScaGWR:
            x0 = jnp.array(model.kernel.params)
        else:
            raise ValueError("Unknown model class")

        x0 = model._to_unconstrained(x0)

        opt_state = self.optax_optim.init(x0)

        if diff_mode == "manual":

            def step(x, opt_state, idx):
                value = model.unconstrained_loss(x, idx)
                grads = model.unconstrained_grad(x, idx)
                updates, opt_state = self.optax_optim.update(grads, opt_state)
                x = optax.apply_updates(x, updates)

                return value, x, opt_state

        elif diff_mode == "auto":

            def step(x, opt_state, idx):
                value, grads = value_and_grad(model.unconstrained_loss, argnums=0)(
                    x, idx
                )

                updates, opt_state = self.optax_optim.update(grads, opt_state)
                x = optax.apply_updates(x, updates)

                return value, x, opt_state

        loss = []
        best_loss = jnp.inf
        count = 0
        x = x0
        with tqdm(total=int(maxiter), disable=not verbose) as pbar:
            for i in range(maxiter):
                pbar.update(1)
                if PRNGkey is None:
                    idx = None
                else:
                    key, PRNGkey = random.split(PRNGkey)
                    idx = random.choice(key, model.N, shape=(batchsize,), replace=True)
                value, x, opt_state = step(x, opt_state, idx)
                loss.append(value)
                pbar.set_description("loss = {:.5f}, raw_params = {}".format(value, x))

                # convergence check
                if value - best_loss < tol:
                    best_loss = float(min(loss))
                    count = 0
                else:
                    count += 1
                    if count >= n_iter_no_change:
                        self.converged = True
                        break
            else:
                self.converged = False

        model.set_params(x)

        return loss


class scipy_optimizer(object):
    def __init__(self) -> None:
        self.xla_jit = True

    def run(
        self, model, method=None, diff_mode="manual", tol=None, kwargs_minimize=dict()
    ):
        diff_mode = diff_mode.lower()
        assert diff_mode in ["manual", "auto", "num"]

        if type(model) is models.GWR_Ridge:
            x0 = jnp.concatenate(
                [jnp.array(model.kernel.params), jnp.array([model.penalty])]
            )
        elif type(model) is models.GWR or type(model) is models.ScaGWR:
            x0 = jnp.array(model.kernel.params)
        else:
            raise ValueError("Unknown model class")

        x0 = model._to_unconstrained(x0)

        def f(x):
            return model.unconstrained_loss(x)

        g = None
        if diff_mode == "manual":

            def g(x):
                return model.unconstrained_grad(x)

        elif diff_mode == "auto":

            def g(x):
                return grad(f)(x)

        if self.xla_jit:
            f = jit(f)
            if g is not None:
                g = jit(g)

        res = optimize.minimize(
            fun=f, x0=x0, method=method, jac=g, tol=tol, **kwargs_minimize
        )

        model.set_params(res.x)

        return res

    def run_scalar(
        self,
        model,
        method="brent",
        bracket=None,
        bounds=None,
        tol=None,
        options=None,
        aicc=False,
        kwargs_minimize=dict(),
    ):
        if aicc:

            def f(x):
                z = model._to_constrained(jnp.array([x]))
                return model.AICc([z])

        else:

            def f(x):
                return model.unconstrained_loss(jnp.array([x]))

        if self.xla_jit:
            f = jit(f)

        res = optimize.minimize_scalar(
            fun=f,
            bracket=bracket,
            bounds=bounds,
            method=method,
            tol=tol,
            options=options,
            **kwargs_minimize
        )

        model.set_params(jnp.array([res.x]))

        return res


class scipy_L_BFGS_B(object):
    """same setting to the 'scgwr' package in R
    see:  https://github.com/cran/scgwr/blob/master/R/scgwr.R
    """

    def __init__(self) -> None:
        self.xla_jit = True

    def run(self, model, diff_mode="auto", tol=None, kwargs_minimize=dict()):

        diff_mode = diff_mode.lower()
        assert diff_mode in ["manual", "auto", "num"]

        if type(model) is models.GWR_Ridge:
            x0 = jnp.concatenate(
                [jnp.array(model.kernel.params), jnp.array([model.penalty])]
            )

            def f(x):
                return model.loocv_loss(x[:-1], x[-1])

        elif type(model) is models.GWR or type(model) is models.ScaGWR:
            x0 = jnp.array(model.kernel.params)

            def f(x):
                return model.loocv_loss(x)

        else:
            raise ValueError("Unknown model class")

        if self.xla_jit:
            f = jit(f)

        def f64(x):
            return np.array(f(x), dtype=np.float64)

        g = None
        if diff_mode == "manual":
            if type(model) is models.GWR_Ridge:

                def g(x):
                    g1 = model.grad_params_loocv(x[:-1], x[-1])
                    g2 = model.grad_penalty_loocv(x[:-1], x[-1])

                    return jnp.concatnate([g1, jnp.array([g2])])

            elif type(model) is models.GWR or type(model) is models.ScaGWR:

                def g(x):
                    return model.grad_params_loocv(x)

        elif diff_mode == "auto":

            def g(x):
                return grad(f)(x)

        if self.xla_jit and g is not None:
            g = jit(g)

        if g is None:
            g64 = None
        else:

            def g64(x):
                return np.array(g(x)).astype(np.float64)

        res = optimize.minimize(
            fun=f64,
            x0=np.array(x0, dtype=np.float64),
            method="L-BFGS-B",
            jac=g64,
            bounds=optimize.Bounds(lb=np.zeros(len(x0)), ub=np.inf * np.ones(len(x0))),
            tol=tol,
            **kwargs_minimize
        )

        model.set_params(res.x, transform=False)

        return res




"""
Optimizers using stochastic gradients with second order informations.
Faster convergence can be expected.
"""
from jax import numpy as jnp
from jax import random

from tqdm.auto import tqdm

from .. import models
from .sg import SGD

__all__ = ["SGN", "SGN_BFGS", "SGN_LM"]


class SGN(SGD):
    """
    Stochastic Gauss-Newton methods.
    Small value is added to approximated Hessian to guarantee its positive definitness.

    refernces:
    Bottou, L., Curtis, F. E., & Nocedal, J. (2018).
    Optimization methods for large-scale machine learning.
    In SIAM Review (Vol. 60, Issue 2, pp. 223–311).
    """

    def __init__(self, learning_rate0=1.0, lam=0):
        super().__init__(learning_rate0=learning_rate0, lam=lam)

        self._delta = 1e-3

    def run(
        self,
        model,
        maxiter=1000,
        batchsize=100,
        PRNGkey=random.PRNGKey(seed=123),
        diff_mode="manual",  # ignored
        tol=0.001,
        n_iter_no_change=100,
        verbose=True,
    ):

        maxiter = int(maxiter)
        batchsize = int(batchsize)
        assert batchsize <= model.N
        x0, f_g_J = self._init_optimizer(model)

        key, PRNGkey = random.split(PRNGkey)
        idx = random.choice(key, model.N, shape=(batchsize,), replace=True)
        l, g, J = f_g_J(x0, idx=idx)
        loss = [float(l)]
        x = x0
        best_loss = float(loss[-1])
        count = 0
        with tqdm(total=maxiter, disable=not verbose) as pbar:
            for t in range(maxiter):
                pbar.update(1)
                pbar.set_description("loss={}".format(loss[-1]))

                self.lr = self.lr_schedule(t + 1)
                key, PRNGkey = random.split(PRNGkey)
                idx = random.choice(key, model.N, shape=(batchsize,), replace=True)
                x, l = self.step(t + 1, x, f_g_J, idx)

                loss.append(l)

                # convergence check
                if l - best_loss < tol:
                    best_loss = float(min(loss))
                    count = 0
                else:
                    count += 1
                    if count >= n_iter_no_change:
                        self.converged = True
                        break
            else:
                self.converged = False

        model.set_params(x)

        return loss

    def _init_optimizer(self, model):
        if type(model) is models.GWR_Ridge:
            x0 = jnp.concatenate(
                [jnp.array(model.kernel.params), jnp.array([model.penalty])]
            )

        elif type(model) is models.GWR or type(model) is models.ScaGWR:
            x0 = jnp.array(model.kernel.params)
        else:
            raise ValueError("Unknown model class")

        x0 = model._to_unconstrained(x0)

        def f_g_J(x, idx):
            return model.unconstrained_GN(x, idx)

        self._dim = len(x0)
        self.H = 2 * jnp.eye(1)
        self.Hinv = 1 / self.H

        return x0, f_g_J

    def step(self, t, x, f_g_J, idx):
        loss, g, J = f_g_J(x, idx)
        p = self._direction(g, J)
        x_new = x + self.lr * p
        return x_new, float(loss)

    def _direction(self, g, J):
        H = jnp.repeat(self.H, len(J)).reshape((-1, 1, 1))
        H_approx = jnp.mean(J[..., None] @ H @ J[:, None, :], axis=0)
        # direct inversion in not stable
        p = jnp.linalg.solve(H_approx + self._delta * jnp.eye(self._dim), -g)
        return p


class SGN_BFGS(SGN):
    """
    Stochastic Gauss-Newton methods.
    BFGS formula is applied to guarantee the positive definiteness of approximated Hessian matrix.
    Note: I recommend small learning rate

    refernces:
    Bottou, L., Curtis, F. E., & Nocedal, J. (2018).
    Optimization methods for large-scale machine learning.
    In SIAM Review (Vol. 60, Issue 2, pp. 223–311).
    """

    def _init_optimizer(self, model):
        ret = super()._init_optimizer(model)
        # initalize BFGS formula
        self._Ginv = jnp.eye(self._dim)

        return ret

    def step(self, t, x, f_g_J, idx):
        loss, g, J = f_g_J(x, idx)
        p = self._direction(g, J)
        s = self.lr * p
        x_new = x + s
        self._bfgs_update(s, J)
        return x_new, float(loss)

    def _direction(self, g, J):
        p = -self._Ginv @ g
        return p

    def _bfgs_update(self, s, J):
        H = jnp.repeat(self.H, len(J)).reshape((-1, 1, 1))
        H_approx = jnp.mean(J[..., None] @ H @ J[:, None, :], axis=0)
        # direct inversion in not stable
        # p = jnp.linalg.solve(H_approx, -g)
        v = H_approx @ s

        # BFGS formula
        I = jnp.eye(self._dim)
        sv = jnp.sum(v * s)
        a = I - v[:, None] @ s[None] / sv
        self._Ginv = a.T @ self._Ginv @ a + s[:, None] @ s[None] / sv


class SGN_LM(SGN):
    """
    Stochastic Gauss-Newton methods with damping of Levenberg-Marquardt (LM) method.
    The damping improves the stability.
    This algorithm refered to as "SMW-GN" in Ren and Goldfarb (2019)
    because they used Sherman-Morrison-Woodbury(SMW) formula.
    SMW formula is not used in this implementation
    because we assume that the number of parameters is not far larger than the mini-batch size.

    refernces:
    Ren, Y., & Goldfarb, D. (2019).
    Efficient Subsampled Gauss-Newton and Natural Gradient Methods for Training Neural Networks.
    https://arxiv.org/abs/1906.02353v1
    """

    def __init__(
        self,
        learning_rate0=1.0,
        lam=0,
        lam_LM0=1.0,
        boost=1.01,
        drop=0.99,
        eps=0.25,
        tau=0.001,
    ):
        super().__init__(learning_rate0=learning_rate0, lam=lam)

        assert tau <= lam_LM0
        assert drop < 1 < boost
        assert 0 < eps < 0.5
        assert 0 < tau

        self.lam_LM0 = float(lam_LM0)
        self.boost = float(boost)
        self.drop = float(drop)
        self.eps = float(eps)
        self.tau = float(tau)

        self.lam_LM = float(self.lam_LM0)
        self._lam_LM = self.lam_LM0 - self.tau

    def step(self, t, x, f_g_J, idx):
        loss, g, J = f_g_J(x, idx)
        self._prev_loss = float(loss)
        if t > 1:
            self._LM(loss, g)
        p = self._direction(g, J)
        self._prev_p = jnp.array(p)
        x_new = x + self.lr * p
        return x_new, float(loss)

    def _direction(self, g, J):
        N2 = len(J)
        Hinv = jnp.repeat(self.Hinv, N2).reshape(-1, 1, 1)
        self.B = jnp.mean(J[..., None] @ Hinv @ J[:, None, :], axis=0)
        p = -jnp.linalg.solve(self.B + self.lam_LM + jnp.eye(self._dim), g)
        return p

    def _LM(self, loss, g):
        def m(p):
            return loss + jnp.sum(g * p) + 0.5 * p @ self.B @ p

        rho = (loss - self._prev_loss) / (loss - m(self._prev_p))

        if rho < self.eps:
            self._lam_LM = self._lam_LM * self.boost
        elif 1 - self.eps < rho:
            self._lam_LM = self._lam_LM * self.drop

        self.lam_LM = self.lam_LM + self.tau



"""
Optimizers using stochastic gradients.
"""
import numpy as np

from tqdm.auto import tqdm

from .. import models

__all__ = ["SGD", "ASGD", "SGDarmijo"]


class SGD(object):
    """
    Stochastic Gradient Descent Algorithm

    reference:
    Bottou, L. (2010). Large-scale machine learning with stochastic gradient descent.
    Proceedings of COMPSTAT 2010 - 19th International Conference on Computational Statistics,
    Keynote, Invited and Contributed Papers, 177–186. https://doi.org/10.1007/978-3-7908-2604-3_16
    """

    def __init__(self, learning_rate0=0.1, lam=1e-4):
        self.learning_rate0 = float(learning_rate0)
        self.lam = float(lam)

    def lr_schedule(self, t):
        return self.learning_rate0 / (1 + self.lam * self.learning_rate0 * t)

    def run(
        self,
        model,
        maxiter=1000,
        batchsize=100,
        rng=np.random.default_rng(123),
        tol=0.001,
        n_iter_no_change=100,
        verbose=True,
    ):
        maxiter = int(maxiter)
        batchsize = int(batchsize)
        assert batchsize <= model.N
        x0, [f, g, f_and_g] = self._init_optimizer(model)

        self.batchsize = batchsize
        self.N = model.N

        idx = rng.choice(model.N, size=(batchsize,), replace=True)
        loss = [float(f(x0, idx=idx))]
        x = x0
        best_loss = float(loss[-1])
        count = 0
        self.lr_log = []
        with tqdm(total=maxiter, disable=not verbose) as pbar:
            for t in range(maxiter):
                pbar.update(1)
                pbar.set_description("loss={}".format(loss[-1]))

                self.lr = self.lr_schedule(t + 1)
                idx = rng.choice(model.N, size=(batchsize,), replace=True)
                x, l = self.step(t + 1, x, f, g, f_and_g, idx)
                self.lr_log.append(float(self.lr))

                loss.append(l)

                # convergence check
                if l - best_loss < tol:
                    best_loss = float(min(loss))
                    count = 0
                else:
                    count += 1
                    if count >= n_iter_no_change:
                        self.converged = True
                        break
            else:
                self.converged = False

        model.set_params(x)

        return loss

    def _init_optimizer(self, model):

        if type(model) is models.GWR_Ridge:
            x0 = np.concatenate(
                [np.array(model.kernel.params), np.array([model.penalty])]
            )

        elif type(model) is models.GWR or type(model) is models.ScaGWR:
            x0 = np.array(model.kernel.params)
        else:
            raise ValueError("Unknown model class")

        x0 = model._to_unconstrained(x0)

        def f(x, idx):
            return model.unconstrained_loss(x, idx)

        def g(x, idx):
            return model.unconstrained_grad(x, idx)

        def f_and_g(x, idx):
            return (
                model.unconstrained_loss(x, idx),
                model.unconstrained_grad(x, idx),
            )

        return x0, [f, g, f_and_g]

    def step(self, t, x, f, g, f_and_g, idx):
        grads = g(x, idx)
        x_new = x - self.lr * grads
        loss = f(x, idx)
        return x_new, float(loss)


class ASGD(SGD):
    """
    Avereaged Stochastic Gradient Descent Algorithm

    reference:
    Bottou, L. (2010). Large-scale machine learning with stochastic gradient descent.
    Proceedings of COMPSTAT 2010 - 19th International Conference on Computational Statistics,
    Keynote, Invited and Contributed Papers, 177–186. https://doi.org/10.1007/978-3-7908-2604-3_16
    """

    def __init__(self, learning_rate0=0.1, lam=1e-4):
        super().__init__(learning_rate0=learning_rate0, lam=lam)

    def lr_schedule(self, t):
        return self.learning_rate0 * (1 + self.lam * self.learning_rate0 * t) ** (-0.75)

    def step(self, t, x, f, g, f_and_g, idx):
        grads = g(self._x_sgd, idx)
        self._x_sgd = self._x_sgd - self.lr * grads
        x_new = (t * x + self._x_sgd) / (t + 1)
        loss = f(x_new, idx)
        return x_new, float(loss)

    def _init_optimizer(self, model):
        x0, [f, g, f_and_g] = super()._init_optimizer(model)
        self._x_sgd = np.array(x0)
        return x0, [f, g, f_and_g]


class SGDarmijo(SGD):
    """
    Stochastic Gradient Descent Algorithm with Armijo Line-search

    reference:
    Vaswani, S., Mishkin, A., Laradji, I., Schmidt, M., Gidel, G., & Lacoste-Julien, S. (2019).
    Painless stochastic gradient: Interpolation, line-search, and convergence rates.
    Advances in neural information processing systems, 32.
    """

    def __init__(
        self,
        learning_rate0=1.0,
        c=0.5,
        ls_decay=0.5,
        reset_decay=2.0,
        search_from_lr0=False,
    ):
        self.learning_rate0 = float(learning_rate0)

        assert c > 0.0
        self.c = float(c)

        assert 0.0 < ls_decay and ls_decay < 1.0
        self.ls_decay = float(ls_decay)

        assert reset_decay >= 1.0
        self.reset_decay = float(reset_decay)

        self.search_from_lr0 = search_from_lr0

        self.lr = float(learning_rate0)

        self._ls_max = 100

    def lr_schedule(self, t):
        # set initial lr on line search
        if self.search_from_lr0:
            lr = self.learning_rate0
        else:
            lr = self.lr * self.reset_decay ** (self.batchsize / self.N)
            # print(f"lr0 = {lr}")
        return lr  # / self.ls_decay

    def step(self, t, x, f, g, f_and_g, idx):
        grads = g(x, idx)
        loss = f(x, idx)
        # line search
        self.lr = self.armijo_search(x, grads, self.lr, f=lambda x: f(x, idx))
        x_new = x - self.lr * grads
        return x_new, float(loss)

    def armijo_search(self, x, grads, lr, f):
        g2 = np.sum(np.square(grads))
        f0 = f(x)
        cond_fun = lambda lr: not self.armijo_cond(f, f0, x, grads, g2, lr)
        body_fun = lambda lr: lr * self.ls_decay

        # print(f"start line search (lr={lr})")
        i = 0
        while cond_fun(lr) or i > self._ls_max:
            # print(f"lr={lr}")
            lr = body_fun(lr)
            i += 1

        lr_searched = lr

        return lr_searched

    def armijo_cond(self, f, f0, x, grads, g2, lr):
        x_new = x - lr * grads
        f_new = f(x_new)
        return f_new <= f0 - self.c * lr * g2



"""
Optimizers using variance reduced stochastic gradients.
They are recommended if you require high-accuracy optimization.
"""
from jax import numpy as jnp
from jax import random, lax

from tqdm.auto import tqdm

from .sg import SGD


__all__ = ["SVRG", "KatyushaXs", "KatyushaXw"]


class SVRG(SGD):
    """
    Stochastic Variance Reduced Gradient with mini-batch

    refernces:
    Johnson, R., & Zhang, T. (2013).
    Accelerating Stochastic Gradient Descent using Predictive Variance Reduction.
    In Advances in Neural Information Processing Systems (Vol. 26).

    Allen-Zhu, Z. (2018).
    Katyusha X: Practical Momentum Method for Stochastic Sum-of-Nonconvex Optimization.
    35th International Conference on Machine Learning, ICML 2018, 1, 284–290.
    """

    def __init__(self, learning_rate0=0.1, lam=0):
        super().__init__(learning_rate0=learning_rate0, lam=lam)

    def run(
        self,
        model,
        max_epoch=100,
        batchsize=None,
        PRNGkey=random.PRNGKey(seed=123),
        diff_mode="manual",
        tol=0.001,
        n_iter_no_change=5,
        min_epoch=5,
        # check_converge_exact=True,
        verbose=True,
        lax_scan=True,
    ):
        # maxiter = int(maxiter)
        max_epoch = int(max_epoch)
        if batchsize is None:
            batchsize = jnp.floor(jnp.sqrt(model.N)).astype(int)
        batchsize = int(batchsize)

        assert batchsize <= model.N

        if batchsize > jnp.sqrt(model.N):
            print(
                "Message: Effective batchsize is smaller than {:.1f} (=sqrt(N))".format(
                    jnp.sqrt(model.N)
                )
            )
        x0, [f, g, f_and_g] = self._init_optimizer(model, diff_mode)
        f_step = self._get_f_step(f, g)

        self.M = int(jnp.ceil(model.N / batchsize))
        # self.max_epoch = int(jnp.ceil(maxiter / self.M))
        # if self.max_epoch < min_epoch:
        #     raise ValueError(
        #         "maxiter is too small: it should be larger than {}(= min_epoch × N)".format(
        #             model.N * min_epoch
        #         )
        #     )

        # loss = [float(f(x0, idx=None))]
        loss = jnp.array([float(f(x0, idx=None))])
        self.exact_loss = []
        x = x0
        self._x0 = x0
        y = jnp.array(x)
        y_old = jnp.array(y)
        best_loss = float(loss[-1])
        count = 0
        # stop_flag = False
        t = 0

        with tqdm(total=max_epoch, disable=not verbose) as pbar:
            for epoch in range(max_epoch):
                x = self._momentum(self._x0, y, y_old, epoch)

                # SVRG^1ep
                self._x0 = jnp.array(x)
                l, self._grad_exact = f_and_g(x, idx=None)
                self.exact_loss.append(float(l))

                # convergence check
                if jnp.abs(best_loss - l) > tol:
                    best_loss = float(min(self.exact_loss))
                    count = 0
                elif epoch + 1 >= min_epoch:
                    count += 1
                    if count >= n_iter_no_change:
                        self.converged = True
                        break

                pbar.update(1)
                pbar.set_description(
                    "epoch {}/{}: loss={:.3f}".format(
                        epoch + 1, max_epoch, self.exact_loss[-1]
                    )
                )

                self.lr = self.lr_schedule(t + 1)
                t += 1
                if lax_scan:
                    key, PRNGkey = random.split(PRNGkey)
                    idxs = random.choice(key, model.N, shape=(self.M, batchsize))

                    x, ls = self.batch_step(f_step, x, idxs)
                    loss = jnp.concatenate([loss, ls])

                else:
                    for m in range(self.M):
                        pbar.set_description(
                            "epoch {}/{} ({}/{}): loss={}".format(
                                epoch + 1,
                                max_epoch,
                                m + 1,
                                self.M,
                                self.exact_loss[-1],
                            )
                        )

                        # self.lr = self.lr_schedule(t + 1)
                        key, PRNGkey = random.split(PRNGkey)
                        idx = random.choice(
                            key, model.N, shape=(batchsize,), replace=True
                        )
                        x, l = self.step(t + 1, x, f, g, f_and_g, idx)

                        loss = jnp.append(loss, l)

                # value to get momentum
                y_old = jnp.array(y)
                y = jnp.array(x)

            else:
                self.converged = False

        model.set_params(x)

        return loss

    def batch_step(self, f_step, x0, idxs):
        x, loss = lax.scan(f_step, init=x0, xs=idxs)
        return x, loss

    def _get_f_step(self, f, g):
        def f_step(carry, idx):
            return self.step(t=0, x=carry, f=f, g=g, f_and_g=None, idx=idx)

        return f_step

    def step(self, t, x, f, g, f_and_g, idx):
        grads = self._grad_exact + g(x, idx) - g(self._x0, idx)
        x_new = x - self.lr * grads
        loss = f(x, idx)
        return x_new, loss

    def _momentum(self, x, y, y_old, k):
        return y


class KatyushaXs(SVRG):
    """
    KatyushaXs algorithm

    refernces:
    Allen-Zhu, Z. (2018).
    Katyusha X: Practical Momentum Method for Stochastic Sum-of-Nonconvex Optimization.
    35th International Conference on Machine Learning, ICML 2018, 1, 284–290.
    """

    def __init__(self, learning_rate0=0.1, lam=0, neg_moment=0.1):
        """
        If neg_moment=0.5, it is equivalent to SVRG
        neg_moment is refered as parameter tau in Allen-Zu (2018)
        """
        assert 0 < neg_moment <= 1
        self.neg_moment = float(neg_moment)
        super().__init__(learning_rate0=learning_rate0, lam=lam)

    def _momentum(self, x, y, y_old, k):
        return (1.5 * y + 0.5 * x - (1 - self.neg_moment) * y_old) / (
            1 + self.neg_moment
        )


class KatyushaXw(SVRG):
    """
    KatyushaXw algorithm

    refernces:
    Allen-Zhu, Z. (2018).
    Katyusha X: Practical Momentum Method for Stochastic Sum-of-Nonconvex Optimization.
    35th International Conference on Machine Learning, ICML 2018, 1, 284–290.
    """

    def _momentum(self, x, y, y_old, k):
        return ((3 * k + 1) * y + (k + 1) * x - (2 * k - 2) * y_old) / (2 * k + 4)




try:
    from jax import numpy as jnp
    from jax import value_and_grad, vmap

    _JAX_AVAILABLE = True
except:
    import numpy as jnp

    print("No JAX mode: use numpy instead")

    def value_and_grad(f):
        raise ValueError("Autograd is not available: Install JAX")

    def vmap(f):
        def vf(idx):
            return jnp.stack(list(map(f, idx)), axis=0)

        return vf

    _JAX_AVAILABLE = False

from . import kernels

from scipy import stats
from tqdm.auto import tqdm
import copy


class GWR_Ridge(object):
    def __init__(self, y, X, sites, kernel=kernels.Gaussian([1.0]), penalty=0.01):
        self.y = jnp.array(y).reshape(-1, 1)
        self.N = len(self.y)
        self.X = jnp.array(X).reshape(self.N, -1)
        self.D = self.X.shape[1]
        self.sites = jnp.array(sites).reshape(self.N, -1)

        assert issubclass(kernel.__class__, kernels._baseKernel)
        self.kernel = kernel

        assert penalty >= 0
        self.penalty = penalty

    def get_beta(self, s):
        return self._get_beta(s, self.kernel.params, self.penalty, loocv=None)

    def _get_beta(self, s, params, penalty, loocv):
        w = self.kernel.forward(s, self.sites, params=params, loocv=loocv)
        P = self.X.T * w.reshape(1, -1)

        beta = jnp.linalg.solve(
            P @ self.X + penalty * jnp.eye(self.D), P @ self.y
        ).flatten()

        return beta

    def set_betas_inner(self):
        self.betas = vmap(self._set_beta_inner)(jnp.arange(self.N, dtype=int))

    def _set_beta_inner(self, i):
        return self._get_beta(
            self.sites[i], self.kernel.params, self.penalty, loocv=None
        )

    def _to_constrained(self, unconstrained):
        # return jnp.exp(unconstrained)
        return jnp.logaddexp(0, unconstrained)  # softplus function

    def _grad_to_constrained(self, unconstrained):
        # return jnp.exp(unconstrained)
        return 0.5 * (
            jnp.tanh(0.5 * unconstrained) + 1
        )  # sigmoid function = derivative of softplus function

    def _to_unconstrained(self, constrained):
        # return jnp.log(constrained)
        return constrained + jnp.log(
            1 - jnp.exp(-constrained)
        )  # inverse of softplus function

    def set_params(self, unconstrained, transform=True):
        if transform:
            z = self._to_constrained(unconstrained)
        else:
            z = unconstrained
        self.kernel.params = z[:-1]
        self.penalty = z[-1]

    def unconstrained_loss(self, x, idx=None):
        z = self._to_constrained(x)
        return self.loocv_loss(z[:-1], z[-1], idx)

    def unconstrained_grad(self, x, idx=None):
        z = self._to_constrained(x)
        gz = self._grad_to_constrained(x)
        g1 = self.grad_params_loocv(z[:-1], z[-1], idx) * gz[:-1]
        g2 = self.grad_penalty_loocv(z[:-1], z[-1], idx) * gz[-1]
        return jnp.concatenate([g1, jnp.array([g2])])

    def loocv_loss(self, params, penality, idx=None):
        def f(i):
            return self._loocv_loss(i, params, penality)

        if idx is None:
            idx = jnp.arange(self.N, dtype=int)

        return jnp.mean(vmap(f)(idx))

    def _loocv_loss(self, i, params, penalty):
        beta = self._get_beta(self.sites[i], params, penalty, loocv=i)
        eps = self.y[i] - jnp.sum(beta * self.X[i])
        return eps**2

    def loocv_GN(self, params, penalty, idx=None):
        def f_g(i):
            return value_and_grad(self._loocv_pred, argnums=[1, 2])(
                i, jnp.array(params), penalty
            )

        if idx is None:
            idx = jnp.arange(self.N, dtype=int)

        pred, J = vmap(f_g)(idx)  # idxが長すぎるとメモリが落ちる
        J = jnp.concatenate(
            [J[0].reshape((len(idx), len(params))), J[1][:, None]], axis=-1
        )

        eps = pred.reshape(-1, 1) - self.y[idx]
        loocv = jnp.mean(jnp.square(eps))
        g = jnp.mean(eps * J, axis=0)
        return loocv, g, J

    def unconstrained_GN(self, x, idx=None):
        z = self._to_constrained(x)
        gz = self._grad_to_constrained(x)
        loocv, g, J = self.loocv_GN(z[:-1], z[-1], idx=idx)
        g = g * gz
        J = J * gz[None]
        return loocv, g, J

    def _loocv_pred(self, i, params, penalty):
        beta = self._get_beta(self.sites[i], params, penalty, loocv=i)
        return jnp.sum(beta * self.X[i])

    def grad_params_loocv(self, params, penalty, idx=None):
        def f(i):
            return self._grad_params_loocv(i, params, penalty)

        if idx is None:
            idx = jnp.arange(self.N, dtype=int)

        return jnp.mean(vmap(f)(idx), axis=0)

    def _grad_params_loocv(self, i, params, penalty):
        w = self.kernel.forward(self.sites[i], self.sites, params=params, loocv=i)
        dw = self.kernel.grad(self.sites[i], self.sites, params=params, loocv=i)

        P = self.X.T * w.reshape(1, -1)

        K = P @ self.X + penalty * jnp.eye(self.D)
        beta = jnp.linalg.solve(K, P @ self.y).flatten()
        eps = self.y - self.X @ beta[:, None]

        dP = self.X.T[None] * dw.T[:, None, :]
        C = dP @ eps[None]
        grad_y = self.X[i, None, None] @ jnp.linalg.solve(K[None], C)

        return -2 * grad_y.flatten() * eps[i]

    def grad_penalty_loocv(self, params, penalty, idx=None):
        def f(i):
            return self._grad_penalty_loocv(i, params, penalty)

        if idx is None:
            idx = jnp.arange(self.N, dtype=int)

        return jnp.mean(vmap(f)(idx))

    def _grad_penalty_loocv(self, i, params, penalty):
        w = self.kernel.forward(self.sites[i], self.sites, params=params, loocv=i)

        P = self.X.T * w.reshape(1, -1)

        K = P @ self.X + penalty * jnp.eye(self.D)
        beta = jnp.linalg.solve(K, P @ self.y).flatten()

        grad_y = -self.X[i] @ jnp.linalg.solve(K, beta)

        eps = self.y[i, 0] - jnp.sum(self.X[i] * beta)

        return -2 * grad_y * eps

    def setInferenceStats(self, alpha=0.05):
        def func(i):
            C = self._getC(i)
            y = self.y.flatten()
            x = self.X[i]
            eps2 = jnp.square(y[i] - x @ C @ y)
            r = x @ C[:, i]
            diagCCt = jnp.diag(C @ C.T)
            beta = C @ y

            return eps2, r, diagCCt, beta

        eps2, r, diagCCt, self.betas = vmap(func)(jnp.arange(self.N, dtype=int))

        self.ENP = jnp.sum(r)
        self.RSS = jnp.sum(eps2)
        self.df = self.N - self.ENP

        # ML estimator (which is the same setting to mgwr package in Python)
        # It is consistent to the original paper of AICc
        # (Hurvich, Simonoff, Tsai 1998, J. R. Statist. Soc. B)
        self.sigma2_ML = self.RSS / self.N
        # unbiased estimator
        self.sigma2 = self.RSS / (self.N - self.ENP)

        self.aicc = self.N * (
            jnp.log(self.sigma2_ML)
            + jnp.log(2 * jnp.pi)
            + (self.N + self.ENP) / (self.N - 2 - self.ENP)
        )

        # local test statistics
        self.tvalues = self.betas / jnp.sqrt(diagCCt * self.sigma2)
        self.alpha = float(alpha)
        cdf = stats.t(self.df).cdf(self.tvalues)
        self.pvalues = 2 * jnp.where(self.tvalues < 0, cdf, 1 - cdf)

        # ENP should be 2tr[H] -tr[H'H] in original paper, but we follows mgwr implementation
        # ENP mentioned above is not implemented currently
        self.alpha_adj = self.alpha / self.ENP * self.D

        self.significant = self.pvalues <= self.alpha_adj

    def _getC(self, i):
        w = self.kernel.forward(
            self.sites[i], self.sites, params=self.kernel.params, loocv=None
        )
        P = self.X.T * w.reshape(1, -1)

        return jnp.linalg.solve(P @ self.X + self.penalty * jnp.eye(self.D), P)

    def AICc(self, params=None, sigma2_type=0):
        """Fast Evaluation of AICc

        reference.
        Li, Z., Fotheringham, A. S., Li, W., & Oshan, T. (2019).
        Fast Geographically Weighted Regression (FastGWR):
        a scalable algorithm to investigate spatial process heterogeneity
        in millions of observations.
        International Journal of Geographical Information Science, 33(1), 155–175.
        """
        if params is not None:
            self.kernel.params = params

        eps2, r = vmap(self._aicciter)(jnp.arange(self.N))
        enp = jnp.sum(r)
        rss = jnp.sum(eps2)

        if sigma2_type != 1:
            # ML estimator (which is the same setting to mgwr package in Python)
            # It is consistent to the original paper of AICc
            # (Hurvich, Simonoff, Tsai 1998, J. R. Statist. Soc. B)
            sigma2 = rss / self.N
        else:
            # unbiased estimator (Li et al. 2019)
            sigma2 = rss / (self.N - enp)

        aicc = self.N * (
            jnp.log(sigma2) + jnp.log(2 * jnp.pi) + (self.N + enp) / (self.N - 2 - enp)
        )
        return aicc

    def _aicciter(self, i):
        Cstar = self._getC(i)
        y = self.y.flatten()
        x = self.X[i]
        eps2 = jnp.square(y[i] - x @ Cstar @ y.flatten())
        r = x @ Cstar[:, i]

        return eps2, r

    def grad_params_aicc(self, params, penalty, idx=None, sigma2_type=0):
        def f(i):
            return self._grad_params_aicc(i, params, penalty)

        idx = jnp.arange(self.N, dtype=int)

        grad_y, grad_r, eps, r = vmap(f)(idx)

        enp = jnp.sum(r)
        denp = jnp.sum(grad_r, axis=0)

        if sigma2_type != 1:
            sigma2 = jnp.mean(jnp.square(eps))
            dsigma2 = -2 * jnp.sum(grad_y * eps[:, None], axis=0) / self.N
        else:
            sigma2 = jnp.sum(jnp.square(eps)) / (self.N - enp)
            dsigma2 = (denp * sigma2 - 2 * jnp.sum(grad_y * eps[:, None], axis=0)) / (
                self.N - enp
            )

        dAICc = (
            self.N / sigma2 * dsigma2
            + 2 * self.N * (self.N - 1) / jnp.square(self.N - 2 - enp) * denp
        )

        return dAICc

    def _grad_params_aicc(self, i, params, penalty):
        w = self.kernel.forward(self.sites[i], self.sites, params=params, loocv=None)
        dw = self.kernel.grad(self.sites[i], self.sites, params=params, loocv=None)

        P = self.X.T * w.reshape(1, -1)

        K = P @ self.X + penalty * jnp.eye(self.D)
        C = jnp.linalg.solve(K, P)
        r = self.X[i] @ C[:, i]
        dC = jnp.linalg.solve(
            K[None],
            self.X.T[None] * dw.T[:, None, :] @ (-self.X @ C + jnp.eye(self.N))[None],
        )
        beta = jnp.linalg.solve(K, P @ self.y).flatten()
        eps = self.y - self.X @ beta[:, None]

        dP = self.X.T[None] * dw.T[:, None, :]
        grad_y = self.X[i, None, :] @ jnp.linalg.solve(K[None], dP @ eps[None])

        grad_r = jnp.sum(self.X[i, None, :] * dC[..., i], axis=-1)

        return (grad_y.reshape((len(params),)), grad_r, eps.flatten()[i], r)

    def grad_penalty_aicc(self, params, penalty, idx=None):
        raise NotImplementedError()


class GWR(GWR_Ridge):
    def __init__(self, y, X, sites, kernel=kernels.Gaussian([1.0])):
        super().__init__(y, X, sites, kernel=kernel, penalty=0.0)

    def loocv_loss(self, params, idx=None):
        return super().loocv_loss(params, penality=self.penalty, idx=idx)

    def loocv_GN(self, params, idx=None):
        def f_g(i):
            return value_and_grad(self._loocv_pred, argnums=1)(
                i, jnp.array(params), penalty=self.penalty
            )

        if idx is None:
            idx = jnp.arange(self.N, dtype=int)

        pred, J = vmap(f_g)(idx)
        J = J.reshape((len(idx), len(params)))

        eps = pred.reshape(-1, 1) - self.y[idx]
        loocv = jnp.mean(jnp.square(eps))
        g = jnp.mean(eps * J, axis=0)
        return loocv, g, J

    def unconstrained_GN(self, x, idx=None):
        z = self._to_constrained(x)
        gz = self._grad_to_constrained(x)
        loocv, g, J = self.loocv_GN(z, idx=idx)
        g = g * gz
        J = J * gz[None]
        return loocv, g, J

    def grad_params_loocv(self, params, idx=None):
        return super().grad_params_loocv(params, penalty=self.penalty, idx=idx)

    def grad_params_aicc(self, params, idx=None, sigma2_type=0):
        return super().grad_params_aicc(
            params, penalty=self.penalty, idx=idx, sigma2_type=sigma2_type
        )

    def unconstrained_loss(self, x, idx=None):
        z = self._to_constrained(x)
        return self.loocv_loss(z, idx)

    def unconstrained_grad(self, x, idx=None):
        z = self._to_constrained(x)
        gz = self._grad_to_constrained(x)
        return self.grad_params_loocv(z, idx) * gz

    def set_params(self, unconstrained, transform=True):
        if transform:
            z = self._to_constrained(unconstrained)
        else:
            z = unconstrained
        self.kernel.params = z


class ScaGWR(GWR):
    def __init__(self, y, X, sites, kernel, precompute=True):
        super().__init__(y, X, sites, kernel=kernel)

        assert type(precompute) is bool
        self.precompute = precompute

        self.M0 = self.X.T @ self.X
        self.m0 = self.X.T @ self.y
        # precomputation
        self._set_active_index()
        if self.precompute:
            if _JAX_AVAILABLE:
                self.m, self.M = vmap(self._getM)(jnp.arange(self.N, dtype=int))
                self.m, self.M = self.m.block_until_ready(), self.M.block_until_ready()
            else:
                self.m = jnp.empty((self.N, self.kernel.n_poly, self.D, 1))
                self.M = jnp.empty((self.N, self.kernel.n_poly, self.D, self.D))

                for i in range(self.N):
                    self.m[i], self.M[i] = self._getM(i)

    def _set_beta_inner(self, i):
        active_idx = jnp.append(self._active_idx[i], i)
        w = self.kernel.forward(
            self.sites[i], self.sites[active_idx], self.kernel.params, loocv=None
        )
        X = self.X[active_idx]
        P = X.T * w.reshape(1, -1)

        beta = jnp.linalg.solve(
            P @ X + self.penalty * jnp.eye(self.D), P @ self.y[active_idx]
        ).flatten()

        return beta

    # def get_beta(self, s):
    #     # knn search does not work with jax.vmap
    #     _, active_idx = self.kernel._knn(s, k=self.kernel.n_neighbour)
    #     w = self.kernel.forward(
    #         s, self.sites[active_idx], self.kernel.params, loocv=None
    #     )
    #     X = self.X[active_idx]
    #     P = X.T * w.reshape(1, -1)

    #     beta = jnp.linalg.solve(
    #         P @ X + self.penalty * jnp.eye(self.D), P @ self.y[active_idx]
    #     ).flatten()

    #     return beta

    def _getM(self, i):
        active_idx = self._active_idx[i]
        # ith weight is excluded from active_idx
        w = self.kernel.base_kernel(self.sites[i], self.sites[active_idx], loocv=None)

        p = jnp.arange(1, self.kernel.n_poly + 1, dtype=int)
        Wp = w[None] ** (4 / jnp.power(2, p))[:, None]  # (p,n)

        X = self.X[active_idx]  # (n,d)
        y = self.y[active_idx]  # (n,1)

        P = X.T[None] * Wp[:, None]  # (1,d,n) * (p,1,n) = (p,d,n)
        M = P @ X[None]  # (p,d,n) @ (1,n,d)
        m = P @ y[None]  # (p,d,n) @ (1,n,1)

        return m, M

    def _set_active_index(self):
        _, idx = self.kernel._knn(self.sites, k=self.kernel.n_neighbour)
        self._active_idx = jnp.array(idx[:, 1:])
        # d = jnp.linalg.norm(self.sites[None] - self.sites[:, None], axis=-1)
        # self._active_idx = jnp.argsort(d, axis=1)[:, 1 : self.kernel.n_neighbour]

    def _loocv_loss(self, i, params, penalty):
        beta = self._get_beta_scagwr(i, params)
        eps = self.y[i] - jnp.sum(beta * self.X[i])
        return eps**2

    def _loocv_pred(self, i, params, penalty):
        beta = self._get_beta_scagwr(i, params)
        return jnp.sum(beta * self.X[i])

    def _get_beta_scagwr(self, i, params):
        (r, R), _, _ = self._get_scagwr_stats(i, params)
        beta = jnp.linalg.solve(R, r)
        return beta.flatten()

    def grad_params_loocv(self, params, idx=None):
        def f(i):
            return self._grad_params_loocv(i, params)

        if idx is None:
            idx = jnp.arange(self.N, dtype=int)

        return jnp.mean(vmap(f)(idx), axis=0)

    def _grad_params_loocv(self, i, params):
        b = params[1:]

        (r, R), (m, M), (m0, M0) = self._get_scagwr_stats(i, params)
        beta = jnp.linalg.solve(R, r).flatten()
        eps = self.y[i] - self.X[i] @ beta

        Rinv = jnp.linalg.inv(R)

        # dbeta/da
        drda = m0.flatten()
        dRda = M0
        dbetada = Rinv @ (-dRda @ beta + drda)

        # dbeta/db
        if len(jnp.array(b)) == 1:
            p = jnp.arange(1, self.kernel.n_poly + 1, dtype=int)
            bp = p * jnp.power(b, p - 1)
            bp = bp[:, None, None]
            drdb = jnp.sum(bp * m, axis=0)[None]
            dRdb = jnp.sum(bp * M, axis=0)[None]
        else:
            drdb = m
            dRdb = M
        dbetadb = Rinv[None] @ (-dRdb @ beta[None, :, None] + drdb)

        grad_beta = jnp.concatenate([dbetada[None, :, None], dbetadb], axis=0)

        grad_y = self.X[i, None, None] @ grad_beta

        return -2 * grad_y.flatten() * eps

    def _get_scagwr_stats(self, i, params):
        a = params[0]
        b = params[1:]

        m0 = self.m0 - (self.X[i] * self.y[i]).reshape(self.m0.shape)
        M0 = self.M0 - self.X[i][:, None] @ self.X[i][None]

        if self.precompute:
            m = self.m[i]
            M = self.M[i]
        else:
            m, M = self._getM(i)

        if len(jnp.array(b)) == 1:
            b = jnp.power(b, jnp.arange(1, self.kernel.n_poly + 1, dtype=int))

        R = a * M0 + jnp.sum(b[:, None, None] * M, axis=0)
        r = a * m0 + jnp.sum(b[:, None, None] * m, axis=0)

        return (r, R), (m, M), (m0, M0)


class MGWR(GWR_Ridge):
    def __init__(
        self,
        y,
        X,
        sites,
        kernel=kernels.Gaussian([1]),
        base_class=GWR_Ridge,
        base_class_params=dict(),
    ):
        super().__init__(y, X, sites)

        if isinstance(kernel, kernels._baseKernel):
            self.kernel = [copy.deepcopy(kernel)] * self.D
        else:
            self.kernel = kernel
            if len(self.kernel) != self.D:
                raise ValueError("kernel should be the iterable object of length D")

        if base_class in (GWR_Ridge, GWR, ScaGWR):
            self.base_class = [base_class] * self.D
        else:
            self.base_class = base_class
            if len(self.base_class) != self.D:
                raise ValueError("base_class should be the iterable object of length D")

        self.base_class_params = base_class_params

    def backfitting(
        self, optimizers, maxiter=100, verbose=True, tol=1e-5, run_params=dict()
    ):

        if len(optimizers) != self.D:
            raise ValueError(
                "optimizers should be the list of optimizer for each exogenous variable"
            )

        # initialize with OLS
        self.betas = jnp.repeat(
            jnp.linalg.solve(a=self.X.T @ self.X, b=(self.X.T @ self.y).flatten())[
                jnp.newaxis
            ],
            self.N,
            axis=0,
        )

        self.pred = self.betas * self.X

        self.RSS = [jnp.sum(jnp.square(self.y - self.pred))]
        self.SOC = [jnp.inf]

        with tqdm(total=maxiter, disable=not verbose) as pbar:
            for i in range(maxiter):
                pbar.update(1)
                pbar.set_description(
                    "RSS = {:.2f}, SOC = {:.2f}%".format(
                        self.RSS[-1], 100 * self.SOC[-1]
                    )
                )

                for d in range(self.D):
                    # local fit
                    resid = self.y - (
                        self.pred.sum(axis=1) - (self.betas[:, d] * self.X[:, d])
                    ).reshape(self.N, 1)
                    localmodel = self.base_class[d](
                        y=resid,
                        X=self.X[:, d].reshape(self.N, 1),
                        sites=self.sites,
                        kernel=self.kernel[d],
                        **self.base_class_params,
                    )

                    optim = optimizers[d]
                    optim.run(localmodel, **run_params)
                    # print(localmodel.kernel.params)
                    self.kernel[d] = copy.deepcopy(localmodel.kernel)

                    # update coefficients
                    localmodel.set_betas_inner()
                    if _JAX_AVAILABLE:
                        self.betas = self.betas.at[:, d].set(localmodel.betas[:, d])
                        self.pred = self.pred.at[:, d].set(
                            self.betas[:, d] * self.X[:, d]
                        )
                    else:
                        self.betas[:, d] = localmodel.betas[:, d]
                        self.pred[:, d] = self.betas[:, d] * self.X[:, d]

                # score of change
                self.RSS.append(jnp.sum(jnp.square(self.y - self.pred)))
                self.SOC.append(jnp.abs(self.RSS[-1] - self.RSS[-2]) / self.RSS[-1])

                if self.SOC[-1] < tol:
                    break

        return self.RSS
